#!/usr/bin/perl

########################################################################
# metaf2xml.pl 1.55
#   example script to parse and print METAR, TAF, SYNOP and BUOY messages
#
# copyright (c) 2006-2014 metaf2xml
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA, or see
# <http://www.gnu.org/licenses/>.
########################################################################

########################################################################
# some things strictly Perl
########################################################################
use strict;
use warnings;
use 5.008_001;

=head1 NAME

metaf2xml.pl

=head1 SYNOPSIS

 metaf2xml.pl [OPTION]... [MESSAGE]...

=head1 DESCRIPTION

This script is an example interface to the Perl module C<metaf2xml::parser>.
Messages can be specified as arguments, each message in single or double quotes.
If no messages are given as arguments but the option -f is given, messages are
read from the specified file. If neither arguments nor option -f are given,
messages are read from the standard input.

The input is expected to consist of one message per line and in the format
specified by the I<WMO Manual No. 306>, B<without modifications due to
distribution> like providing the initial part of messages only once for
several messages or appending an "=" (equal sign) to terminate a message,
and parts of the WMO header prepended according to the option -H.
The program C<metafsrc2raw.pl> can be used to create messages with the required
format from files provided by various public Internet servers.

=head1 DEPENDENCIES

The Perl module C<metaf2xml::parser> is required.

=head1 OPTIONS

=over

=item -v

print version of metaf2xml.pl and exit

=item -f I<msg_file>

read input from I<msg_file>

=item -o I<out_file>

enables writing the data to I<out_file>

=item -H I<0..5>

which parts of the WMO header (C<TTAA[II] CCCC DDHHMM [BBB]>) are
prepended to each message (see L<metafsrc2raw>(1) for an
explanation of its parts):

=over

=item -H0 (default)

(no header)

=item -H1

C<TT> report type

=item -H2

C<TTAA[II]> (like C<-H1> plus region for the stations and optional report number)

=item -H3

C<TTAA[II] CCCC> (like C<-H2> plus report dissimination location)

=item -H4

C<TTAA[II] CCCC DDHHMM> (like C<-H3> plus report dissimination time)

=item -H5

C<TTAA[II] CCCC DDHHMM [BBB]> (like C<-H4> plus optional indicator if the report is delayed/corrected/amended/segmented)

=back

=item -T {C<SPECI>|C<TAF>}

(initial) default message type (default: C<METAR>).
An input line consisting only of one of the keywords C<METAR>, C<SPECI> or
C<TAF> changes the default message type for all following messages.
If the option -H1, -H2, -H3, -H4 or -H5 is given, the message type for each
message is taken from the WMO header.
If a message starts with C<METAR>, C<SPECI> or C<TAF>, this is used as message
type for this message.
If a message starts with C<AAXX>, C<BBXX> or C<OOXX>, the message type SYNOP
is used for this message.
If a message starts with C<ZZYY>, the message type BUOY is used for this
message.

=item -D

include DOCTYPE and reference to the DTD

=item -S I<xslt_file>

include reference to the stylesheet I<xslt_file>

=item -O I<options>

include I<options> (a space separated list of options)

=back

Without the option -o, no output is generated.

=head1 EXAMPLES

Parse a METAR message and write the XML to the standard output:

 metaf2xml.pl -o- "YUDO 090600Z 00000KT CAVOK 22/15 Q1021 NOSIG"

or the same for a SYNOP message:

 metaf2xml.pl -o- "AAXX 17124 74486 32566 63616 10067 20022 39972"

Parse several METAR, TAF, SYNOP and BUOY messages and write the XML to the file
F<example.xml>:

 metaf2xml.pl -o example.xml << EOF
METAR
SBGL 172300Z 24007KT 9999 TS FEW015 FEW030CB BKN100 25/21 Q1013
TAF
SBGL 172130Z 1800/1906 21010KT 7000 BKN015 TX26/1800Z TN17/1823Z PROB30 TEMPO 1801/1805 4000 TSRA BKN015 FEW025CB BECMG 1808/1810 4000 BR BKN008 BECMG 1812/1814 19010KT 6000 NSW BKN012 RMK PGZ
METAR
KJFK 172351Z 14003KT 10SM CLR 14/05 A3036 RMK AO2 SLP280 T01440050 10189 20144 55002
TAF
KJFK 172320Z 1800/1906 VRB04KT P6SM SKC FM181600 18008KT P6SM SKC FM190200 VRB05KT P6SM SKC
METAR
RJTT 172330Z 35009KT CAVOK 22/16 Q1020 NOSIG RMK A3013
TAF
RJTT 172036Z 1721/1824 04012KT 9999 FEW030 BECMG 1806/1809 07014KT BECMG 1812/1815 03006KT
AAXX 17184 74486 NIL
AAXX 17184 47662 12970 20203 10203 20151 30150 40193 56005 60001 80002
ZZYY 31262 17093 13001 522980 042100 6112/ 11119 02801 10021 48214 444 20120 17093 1159/ 518// 60371
ZZYY 41955 17093 2200/ 739065 071940 6111/ 22219 00238 444 20110 17093 1828/ 50101 80119 80007 9/015
ZZYY 53904 17093 2336/ 135184 142953 6112/ 22219 00285 444 20120 17093 2331/ 50101 80092 80016 9/015
EOF

=head1 SEE ALSO

L<metaf2xml::parser>(3pm), L<metafsrc2raw>(1), L<metaf>(1),
L<http://metaf2xml.sourceforge.net/>

=head1 COPYRIGHT and LICENSE

copyright (c) 2006-2014 metaf2xml @ L<http://metaf2xml.sourceforge.net/>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA, or see
L<http://www.gnu.org/licenses/>.

=cut

use Getopt::Std;

our $VERSION = '1.55'; # for Getopt::Std

# path may be changed by install
use lib '/usr/local/lib/perl5/site_perl/5.16'; # METAF2XML_LIB
use metaf2xml::parser 1.055 qw(start_cb start_xml parse_report finish);

########################################################################
# process input and/or pass to parse_report()
########################################################################
{
    my $default_msg_type = 'METAR';

    sub go {
        my $opts_h = shift;
        my ($msg_type, %WMOheader);

        if ($_ eq 'METAR' || $_ eq 'SPECI' || $_ eq 'TAF') {
            $default_msg_type = $_;
            return 1;
        }
        return 1 if $_ eq 'SYNOP' || $_ eq 'BUOY'; # TODO: obsolete

        $msg_type = $default_msg_type;
        if ($opts_h > 0) {
            s/^(..) ?//
                or return 1;
            $WMOheader{TT} = $1;
            $msg_type = { SA => 'METAR',
                          SP => 'SPECI',
                          FC => 'TAF',
                          FT => 'TAF',
                          SM => 'SYNOP',
                          SI => 'SYNOP',
                          SN => 'SYNOP',
                          SS => 'BUOY',
                        }->{$WMOheader{TT}}
                if $WMOheader{TT} =~ /^(?:S[APMINS]|F[CT])$/;

            if ($opts_h >= 2 && s/^([^ ]+) //) {
                $WMOheader{AAII} = $1;
                if ($opts_h >= 3 && s/^([^ ]+) //) {
                    $WMOheader{CCCC} = $1;
                    if ($opts_h >= 4 && s/^(\d\d)(\d\d)(\d\d) //) {
                        @WMOheader{qw(day hour minute)} = ($1, $2, $3);
                        if ($opts_h == 5) {
                            $WMOheader{indicator} = $1
                                if s/^(COR|(?:RR|CC|AA)[A-Z]|P[A-Z]{2}) //;
                        }
                    }
                }
            }
        }

        return parse_report $_, $msg_type;
    }
}

########################################################################
# usage
########################################################################
sub usage {
    print STDERR (shift), <<"EOF";
Usage: metaf2xml.pl [OPTION]... [MESSAGE]...
  Options:
  -v             print version of metaf2xml.pl and exit
  -f <msg_file>  read input from <msg_file>
  -o <out_file>  enables writing the data to <out_file>
  -H <0..5>      which parts of the WMO header are prepended
  -T <msg_type>  (initial) default message type: SPECI, TAF
  -D             include DOCTYPE and reference to the DTD
  -S <xslt_file> include reference to the stylesheet <xslt_file>
  -O <options>   include <options> (a space separated list of options)
Without option -o, no output is generated.

Options may be merged together.  -- stops processing of options.
Space is not required between options and their arguments.

For more details run
        perldoc -F $0
EOF
    exit 1;
}

sub HELP_MESSAGE { usage ''; return; } # for Getopt::Std

########################################################################
# example callback function to print certain values from METAR in a
# different XML format
# Data is cached (in a hash) and printed only at end of report
########################################################################
{
    # CHANGE_HERE: add/delete paths as needed
    my %paths = (
        metar        => '/data/reports/metar/s',
        station      => '/data/reports/metar/obsStationId/s',
        day          => '/data/reports/metar/obsTime/timeAt/day/v',
        hour         => '/data/reports/metar/obsTime/timeAt/hour/v',
        minute       => '/data/reports/metar/obsTime/timeAt/minute/v',
        cloudBase_v_ => '/data/reports/metar/cloud/cloudBase/v',
        airtemp_v    => '/data/reports/metar/temperature/air/temp/v',
        airtemp_u    => '/data/reports/metar/temperature/air/temp/u',
        dewtemp_v    => '/data/reports/metar/temperature/dewpoint/temp/v',
        dewtemp_u    => '/data/reports/metar/temperature/dewpoint/temp/u',
    );

    my (%report, %val);

    # make complete timestamp from day (optional!), hour, minute
    sub date2ts {
        my ($day, $hour, $minute) = @_;

        # CHANGE_HERE: to be implemented
        return (defined $day ? $day : '') . $hour . $minute;
    }

    # convert temperature to C
    sub temp2C {
        my ($val, $unit) = @_;

        return   !defined $val ? ''
               : $unit eq 'C'  ? $val
               :                 sprintf '%.1f', ($val - 32) / 1.8;
    }

    sub printXML2 {
        my $path = shift;
        my $type = shift;
        my $tag  = shift;
        my ($full_path, $jj);

        $full_path = join '/', @$path, $tag;

        # collect data to print
        for (keys %paths) {
            my ($p, $is_eq);

            $p     = $paths{$_};
            $is_eq = $full_path eq $p;

            if ($type eq 'end') {
                if ($is_eq) {
                    if (/_$/) {
                        push @{$report{$_}}, $val{$_};
                    } else {
                        $report{$_} = $val{$_};
                    }
                }
                next;
            }

            $val{$_} = ''
                if $is_eq;

            # check all attributes
            for ($jj = 0; $jj <= $#_; $jj += 2) {
                my ($name, $value) = @_[$jj .. $jj + 1];

                next
                    unless "$full_path/$name" =~ m@^$p(?:/(.*))?$@;

                if (!defined $1) {
                    if (/_$/) {
                        push @{$report{$_}}, $value;
                    } else {
                        $report{$_} = $value;
                    }
                } else {
                    $val{$_} .= " $1:$value";
                }
            }

            # add matching subnode which is empty and has no attributes
            $val{$_} .= " $1"
                if $type eq 'empty' && $#_ == -1 && $full_path =~ m@^$p/(.*)$@;

            if ($type eq 'empty' && $is_eq) {
                if (/_$/) {
                    push @{$report{$_}}, $val{$_};
                } else {
                    $report{$_} = $val{$_};
                }
            }
        }

        # if this were the only callback in this script, the following would
        # simply be done after parse_report()

        # if end of report: print selected data
        if ($#$path == 2 && $type eq 'end') {
            # don't print unless mandatory data is complete
            return 1
                unless exists $report{station} && exists $report{hour};

            # CHANGE_HERE: print added paths/omit printing of deleted paths

            # start of metar, timestamp, station
            print OUT
                    '<metar>',
                    '<date>', date2ts(@report{qw(day hour minute)}), '</date>',
                    '<station>', $report{station}, '</station>'
                or return 0;

            # print clouds (or empty string)
            for (0 .. 2) {
                print OUT "<cloudBase$_>"
                    or return 0;

                if (exists $report{cloudBase_v_}[$_]) {
                    print OUT $report{cloudBase_v_}[$_]
                        or return 0;
                }

                print OUT "</cloudBase$_>"
                    or return 0;
            }

            # print temperatures
            print OUT
                    '<air>', temp2C(@report{qw(airtemp_v airtemp_u)}), '</air>',
                    '<dew>', temp2C(@report{qw(dewtemp_v dewtemp_u)}), '</dew>'
                or return 0;

            # escape special XML characters and print message
            $report{metar} =~ s/&/&amp;/g; # must be the first substitution
            $report{metar} =~ s/</&lt;/g;
            print OUT '<msg>', $report{metar}, '</msg>'
                or return 0;

            # end of metar
            print OUT "</metar>\n"
                or return 0;

            return 1;
        }

        # if top node is closed: close output stream
        close OUT or return 0
            if $#$path == 0 && $type eq 'end';

        return 1;
    }
}

########################################################################
# example callback function to print certain values from reports which
# match some filter criterium
# Data is cached (in a string) and printed only at end of report and
#   if criterium matches
########################################################################
{
    my @nodes2print = (
        '/data/reports/metar/s',
        '/data/reports/metar/ERROR',
        '/data/reports/metar/warning',
        '/data/reports/metar/remark/notRecognised/s',
    );
    my @nodes2filter = (
        '/data/reports/metar/ERROR',
        '/data/reports/metar/warning/warningType',
        '/data/reports/metar/remark/notRecognised',
    );
    my (@print_val, @filter_val, $found, $data_str);

    sub filterVal {
        my $ii = shift;

        if ($ii == 0 || $ii == 2) {
            $found = 1;
        } elsif ($ii == 1) {
            my $value = shift;
            $found = 1
                if $value !~ /msgModified|windMissing|visibilityMissing|tempMissing|QNHMissing/;
        }
        return;
    }

    sub printVal {
        my $path = shift;
        my $type = shift;
        my $tag  = shift;
        my ($full_path, $jj);

        # start of report: init
        if ($#$path == 2 && $type ne 'end') {
            $found    = $#nodes2filter == -1;
            $data_str = '';
        }

        $full_path = join '/', @$path, $tag;

        # collect data to print
        for my $ii (0 .. $#nodes2print) {
            my ($p, $is_eq);

            $p     = $nodes2print[$ii];
            $is_eq = $full_path eq $p;

            if ($type eq 'end') {
                $data_str .= "$ii $p:" . $print_val[$ii] . "\n"
                    if $is_eq;
                next;
            }

            if ($is_eq) {
                $print_val[$ii]  = '';
                $filter_val[$ii] = ();
            }

            # check all attributes
            for ($jj = 0; $jj <= $#_; $jj += 2) {
                my ($name, $value) = @_[$jj .. $jj + 1];

                next
                    unless "$full_path/$name" =~ m@^$p(?:/(.*))?$@;

                if (!defined $1) {
                    $data_str .= "$ii $p:$value\n";
                } else {
                    $print_val[$ii] .= " $1:$value";
                }
            }

            # add matching subnode which is empty and has no attributes
            $print_val[$ii] .= " $1"
                if $type eq 'empty' && $#_ == -1 && $full_path =~ m@^$p/(.*)$@;

            $data_str .= "$ii $p:" . $print_val[$ii] . "\n"
                if $type eq 'empty' && $is_eq;
        }

        # check filters to decide whether to print the report
        for my $ii (0 .. $#nodes2filter) {
            my ($p, $is_eq);

            $p     = $nodes2filter[$ii];
            $is_eq = $full_path eq $p;

            if ($type eq 'end') {
                filterVal $ii, $filter_val[$ii]
                    if $is_eq;
                next;
            }

            # check all attributes
            for ($jj = 0; $jj <= $#_; $jj += 2) {
                my ($name, $value) = @_[$jj .. $jj + 1];

                next
                    unless "$full_path/$name" =~ m@^$p(?:/(.*))?$@;

                if (!defined $1) {
                    filterVal $ii, $value;
                } else {
                    push @{$filter_val[$ii]}, $1, $value;
                }
            }

            # add matching subnode which is empty and has no attributes
            push @{$filter_val[$ii]}, $1, undef
                if $type eq 'empty' && $#_ == -1 && $full_path =~ m@^$p/(.*)$@;

            filterVal $ii, $filter_val[$ii]
                if $type eq 'empty' && $is_eq;
        }

        # if this were the only callback in this script, the following would
        # simply be done after parse_report()

        # if end of report: print selected data unless report is filtered out
        if ($#$path == 2 && $type eq 'end' && $found) {
            print OUT $data_str
                or return 0;
        }

        # if top node is closed: close output stream
        close OUT or return 0
            if $#$path == 0 && $type eq 'end';

        return 1;
    }
}

########################################################################
# example callback function to print node paths with (their) values
# features:
# - Print full path and attributes, but skip nodes with subnodes but
#   without attributes, or only 's' (except for nodes 'ERROR', 'warning'
#   and at the start of the report).
# - Print data in abridged form if a node has only 'v', or 'v' and 'u'.
# - Print line '-/path/...-' to delimit adjacent nodes with the same name.
# - After each report an empty line is printed.
# - Data is not cached, but printed immediately
########################################################################
{
    my $prev_path = '';

    sub simpleDump {
        my $path = shift;
        my $type = shift;
        my $tag  = shift;

        if ($type eq 'end') {
            if ($#$path == 2) {
                $prev_path = '';
                return print OUT "\n";
            }
            $prev_path = join('/', @$path, $tag);
            return 1;
        }

        print OUT "-$prev_path-\n" if join('/', @$path, $tag) eq $prev_path;

        # also delimit adjacent empty nodes with the same name
        $prev_path = join('/', @$path, $tag) if $type eq 'empty';

        return 1 if    $type eq 'start'
                    && $#$path != 2
                    && ($#_ == -1 || ($#_ == 1 && $_[0] eq 's'));

        print OUT join '/', @$path, $tag;

        # print short form for 'v' only, or 'v' and 'u', only
        return print OUT ' = ' . $_[1] . "\n"
            if $#_ == 1 && $_[0] eq 'v';
        return print OUT ' = ' . $_[1] . ' ' . $_[3] . "\n"
            if $#_ == 3 && $_[0] eq 'v' && $_[2] eq 'u';

        # print name and value of attributes (except 's' in most cases)
        while (defined (my $name = shift)) {
            my $value = shift;

            print OUT " $name=$value" or return 0
                if    $name ne 's'
                   || $tag =~ /^(?:metar|taf|synop|buoy|warning|ERROR|notRecognised)$/;
        }
        print OUT "\n";

        # if top node is closed: close output stream
        close OUT or return 0
            if $#$path == 0 && $type eq 'end';

        return 1;
    }
}

########################################################################
# pass all kinds of input to go()
########################################################################
sub main {
    my %opts;

    # get and check options
    $Getopt::Std::STANDARD_HELP_VERSION = 1;
    usage '' unless getopts('vf:o:H:T:DS:O:', \%opts);
    if (exists $opts{v}) {
        print "metaf2xml.pl version $VERSION\n";
        return 1;
    }
    $opts{H} = 0 unless exists $opts{H} && $opts{H} =~ /^[0-5]$/;

    # overwrite initial state for message type
    if (exists $opts{T}) {
        if ($opts{T} eq 'SPECI' || $opts{T} eq 'TAF') {
            $_ = $opts{T};
            go;
        } elsif ($opts{T} eq 'SYNOP' || $opts{T} eq 'BUOY') {
            # TODO: obsolete
        } else {
            usage "ERROR: invalid message type: '" . $opts{T} . "'.\n";
        }
    }

    #open(OUT, ">$opts{o}") && start_cb \&printVal
    #open(OUT, ">$opts{o}") && start_cb \&printXML2
    #open(OUT, ">$opts{o}") && start_cb \&simpleDump
    start_xml \%opts
            or return finish
        if exists $opts{o};

    if ($#ARGV >= 0) {
        RET: for (@ARGV) {
            for (split /\n/, $_) {
                go $opts{H} or last RET;
            }
        }
    } elsif (exists $opts{f}) {
        open my $fh, '<', $opts{f}
            or usage "ERROR: cannot open '$opts{f}'\n";
        while (<$fh>) {
            chomp();
            go $opts{H} or last;
        }
        close $fh;
    } else {
        while (<>) {
            chomp();
            go $opts{H} or last;
        }
    }

    return exists $opts{o} ? finish : 1;
}

exit !main;
