#!/usr/bin/perl

########################################################################
# metafsrc2raw.pl 1.55
#   example script to convert data from different sources to standard format
#
# copyright (c) 2009-2014 metaf2xml
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA, or see
# <http://www.gnu.org/licenses/>.
########################################################################

########################################################################
# some things strictly Perl
########################################################################
use strict;
use warnings;
use 5.008_001;

=head1 NAME

metafsrc2raw.pl

=head1 SYNOPSIS

 metafsrc2raw.pl -F<format> [OPTION]... [FILE]...

=head1 DESCRIPTION

This script is an example interface to the Perl module C<metaf2xml::src2raw>.
It converts METAR, TAF, SYNOP or BUOY information provided in files on various
public Internet servers to messages delimited by a newline and in the format
specified by the I<WMO Manual No. 306>, without modifications due to
distribution like providing the initial part of messages only once for
several messages, or appending an "=" (equal sign) to terminate a message.
It prepends parts of the WMO header according to the option -H.
The output can then be used by the module C<metaf2xml::parser>.

If no input files are specified, then the standard input is read.

=head1 ABBREVIATIONS

=over

=item noaa

National Oceanic and Atmospheric Administration

=item nws

the National Weather Service department of the NOAA

=item iws

the Internet Weather Source of the NWS

=item adds

the Aviation Digital Data Service of the NWS

=item addsds

the experimental data server of the Aviation Digital Data Service of the NWS

=item cod

College of DuPage

=item fsu

Florida State University

=back

=head1 OPTIONS

=over

=item -v

print version of metafsrc2raw.pl and exit

=item -F I<format>

origin of the input data: metar_cycle_iws, staf_cycle_iws, taf_cycle_iws, metaf_nws, metaf_cod, metaf_fsu, synop_nws, synop_cod, synop_fsu, buoy_nws

=item -D

suppress duplicate messages

=item -N

suppress messages that only contain C<NIL>

=item -H I<0..5>

which parts of the WMO header (C<TTAA[II] CCCC DDHHMM [BBB]>) to prepend
to each message:

=over

=item -H0

(no header) (default)

=item -H1

C<TT> (report type:

=over

=item SA

METAR

=item SP

SPECI

=item FT

TAF with forecast period >= 12 hours

=item FC

TAF with forecast period < 12 hours

=item SM

SYNOP at main hours (00:00, 06:00, 12:00, 18:00 UTC)

=item SI

SYNOP at intermediate (main + 3) hours

=item SN

SYNOP at non-standard (other than main and intermediate) hours

=item SS

BUOY

=back

)

=item -H2

C<TTAA[II]> (like C<-H1> plus region for the stations and optional report number)

=item -H3

C<TTAA[II] CCCC> (like C<-H2> plus report dissimination location)

=item -H4

C<TTAA[II] CCCC DDHHMM> (like C<-H3> plus report dissimination time)

=item -H5

C<TTAA[II] CCCC DDHHMM [BBB]> (like C<-H4> plus optional indicator if the report is delayed/corrected/amended/segmented)

=back

=back

=head1 EXAMPLES

Parse SYNOP messages from NOAA/NWS, suppressing NIL messages and duplicates:

 metafsrc2raw.pl -DNF synop_nws << "EOF"
####018000228####
SMPY01 SGAS 060600
AAXX 06064
86134 32970 02402 10250 20235 40105 56005=
86297 NIL=
86134 32970 02402 10250 20235 40105 56005=
86218 32970 01810 10240 20209 40103 58011=
####018000158####
SMFG01 SOCA 060600
AAXX 06064
81405 01259 60505 10241 20229 30105 40114 60071 76086 86200
  333 32024 55048 60067 58002 70098 82705 86813
  555 69905=####018000118####
SMCS01 MROC 060600
AAXX 06064
78762 32970 00915 10220 20160 39094 40101 57005
333 10280 20193 30/// 59006=
EOF

results in:

 AAXX 06064 86134 32970 02402 10250 20235 40105 56005
 AAXX 06064 86218 32970 01810 10240 20209 40103 58011
 AAXX 06064 81405 01259 60505 10241 20229 30105 40114 60071 76086 86200 333 32024 55048 60067 58002 70098 82705 86813 555 69905
 AAXX 06064 78762 32970 00915 10220 20160 39094 40101 57005 333 10280 20193 30/// 59006

Parse METAR/SPECI messages from the College of DuPage with the WMO header
C<TTAAII> prepended:

 metafsrc2raw.pl -F metaf_cod -H2 << "EOF"
355 
SAUS70 KWBC 161200
METAR
KDZJ 161146Z AUTO 00000KT 10SM -DZ SCT002 BKN075 OVC100 09/09
     A3013 RMK AO2=
KLDJ 161150Z AUTO 00000KT 10SM CLR 03/M01 A3027 RMK AO2=
KMLJ 161145Z AUTO 00000KT 4SM RA BKN005 BKN012 OVC033 12/12 A3017
     RMK AO2=
351 
SPCN51 CWAO 161254
SPECI CWCL 161254Z AUTO 19008KT 9SM FEW008 FEW015 BKN027 BKN043
M04/M06 A2947=
EOF

results in:

 SAUS70 METAR KDZJ 161146Z AUTO 00000KT 10SM -DZ SCT002 BKN075 OVC100 09/09 A3013 RMK AO2
 SAUS70 METAR KLDJ 161150Z AUTO 00000KT 10SM CLR 03/M01 A3027 RMK AO2
 SAUS70 METAR KMLJ 161145Z AUTO 00000KT 4SM RA BKN005 BKN012 OVC033 12/12 A3017 RMK AO2
 SPCN51 SPECI CWCL 161254Z AUTO 19008KT 9SM FEW008 FEW015 BKN027 BKN043 M04/M06 A2947

=head1 SEE ALSO

L<metaf2xml::src2raw>(3pm), L<metaf2xml::parser>(3pm),
L<http://metaf2xml.sourceforge.net/>

=head1 COPYRIGHT and LICENSE

copyright (c) 2009-2014 metaf2xml @ L<http://metaf2xml.sourceforge.net/>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA, or see
L<http://www.gnu.org/licenses/>.

=cut

use Getopt::Std;

our $VERSION = '1.55'; # for Getopt::Std

# path may be changed by install
use lib '/usr/local/lib/perl5/site_perl/5.16'; # METAF2XML_LIB
use metaf2xml::src2raw 1.055 qw(start_file next_line);

sub usage {
    print STDERR (shift), <<"EOF";
Usage: metafsrc2raw.pl -F <format> [OPTION]... [FILE] ...
  -v           print version of metafsrc2raw.pl and exit
  -F <format>  format of the input data:
               metar_cycle_iws, staf_cycle_iws, taf_cycle_iws,
               metaf_nws, metaf_cod, metaf_fsu,
               synop_nws, synop_cod, synop_fsu, buoy_nws
  Options:
  -D           suppress duplicate messages
  -N           suppress NIL messages
  -H <0..5>    which parts of the WMO header to include

Options may be merged together.  -- stops processing of options.
Space is not required between options and their arguments.

For more details run
        perldoc -F $0
EOF
    exit 1;
}

sub HELP_MESSAGE { usage ''; return; } # for Getopt::Std

sub main {
    my (%opts, $rc, @msgs);

    # get and check options
    $Getopt::Std::STANDARD_HELP_VERSION = 1;
    usage '' unless getopts('vF:DNH:', \%opts);
    if (exists $opts{v}) {
        print "metafsrc2raw.pl version $VERSION\n";
        return 1;
    }
    usage "ERROR: input format must be specified with -F<FORMAT>.\n"
        unless exists $opts{F};
    $opts{H} = 0 unless exists $opts{H} && $opts{H} =~ /^[0-5]$/;

    $rc = start_file($opts{F}, exists $opts{N}, $opts{H});
    usage "$rc\n" if $rc;
    @msgs = ();
    while (1) {
        $_ = <>;
        for (next_line($_)) {
            if (exists $opts{D}) {
                my $line = $_;
                next if grep { $_ eq $line } @msgs;
                push @msgs, $_;
            }
            print "$_\n";
        }
        last unless defined $_;
    }

    return 1;
}

exit !main;
